"""
Matplotlib specific plotting
"""
import rna

import numpy as np
import re
import warnings
import os
import matplotlib as mpl
import matplotlib.ticker
import matplotlib.pyplot as plt
from matplotlib.patches import Circle
import mpl_toolkits.mplot3d as plt3D
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.dates as dates
from matplotlib.patches import Rectangle
import matplotlib.animation as manimation
from contextlib import contextmanager
from functools import partial
import logging


def show():
    plt.show()


def gca(dim=None, **kwargs):
    """
    Forwarding to plt.gca but translating the dimension to projection
    correct dimension

    Examples:
        >>> ax_initial = rna.plotting.gca()
        >>> del ax_initial

        >>> ax_fresh_implicit = rna.plotting.gca()
        >>> assert axes_dim(ax_fresh_implicit) == 2
        >>> del ax_fresh_implicit

        >>> ax_2 = rna.plotting.gca(2)
        >>> assert axes_dim(ax_2) == 2
        >>> ax_2_implicit = rna.plotting.gca()
        >>> assert axes_dim(ax_2_implicit) == 2
        >>> assert ax_2 is ax_2_implicit

        >>> ax_3 = rna.plotting.gca(3)
        >>> assert axes_dim(ax_3) == 3
        >>> ax_3_implicit = rna.plotting.gca()
        >>> assert axes_dim(ax_3_implicit) == 3
        >>> assert ax_3 is ax_3_implicit

        >>> ax_2 = rna.plotting.gca(2)
        >>> assert axes_dim(ax_2) == 2

    """
    if dim is not None and "projection" not in kwargs:
        if dim == 2:
            kwargs["projection"] = "rectilinear"
            # the line above would already be good ...
            # but there is a bug in matplotlib which ignores setting the default
            axes = plt.gca(**kwargs)
            if axes_dim(axes) != dim:
                return plt.gcf().add_subplot(1, 1, 1, **kwargs)
            else:
                return axes
        elif dim == 3:
            kwargs["projection"] = "3d"
        else:
            raise NotImplementedError(
                "Dimension {dim} not implemented".format(**locals())
            )
    axes = plt.gca(**kwargs)
    return axes


def clear(axes):
    axes.clear()


def upgrade_style(style, source, dest=None):
    """
    Copy a style file at <origionalFilePath> to the <dest> which is the
    foreseen local matplotlib rc dir by default
    The style will be name <style>.mplstyle
    Args:
        style (str): name of style
        source (str): full path to mplstyle file to use
        dest (str): local directory to copy the file to. Matpotlib has to
            search this directory for mplstyle files!
    Examples:
        >>> import rna
        >>> import os
        >>> rna.plotting.use('matplotlib')
        >>> rna.plotting._backend_mod.upgrade_style(
        ...     'rna',
        ...     os.path.join(os.path.dirname(rna.plotting.__file__),
        ...                  'mplstyles/rna.mplstyle'))

    """
    if dest is None:
        dest = mpl.get_configdir()
    style_extension = "mplstyle"
    path = rna.path.resolve(os.path.join(dest, style + "." + style_extension))
    source = rna.path.resolve(source)
    rna.path.cp(source, path)


def set_style(style="rna", dest=None, upgrade=False):
    """
    Set the matplotlib style of name
    Important:
        Either you
    Args:
        style (str)
        dest (str): local directory to use file from. if None, use default
            maplotlib destination
    """
    if dest is None:
        dest = mpl.get_configdir()

    style_extension = "mplstyle"
    path = rna.path.resolve(os.path.join(dest, style + "." + style_extension))
    if style in mpl.style.available and not upgrade:
        plt.style.use(style)
    elif os.path.exists(path) and not upgrade:
        plt.style.use(path)
    else:
        log = logging.getLogger()
        source = rna.path.resolve(
            os.path.join(
                os.path.join(os.path.dirname(__file__), "mplstyles"),
                style + "." + style_extension,
            )
        )
        if os.path.exists(source):
            log.warning("I will copy the style {style} to {dest}.".format(**locals()))
            try:
                upgrade_style(style, source, dest)
                set_style(style)
            except Exception:
                log.error("Could not set style")
        else:
            log.error(
                "Could not set style {path}. Probably you would want to"
                "call rna.plotting.upgrade_style(<style>, "
                "<path to mplstyle file that should be copied>)"
                "once".format(**locals())
            )


def _save(path, **kwargs):
    """
    Args:
        path
        **kwargs:
            width (float): figure width in pts.
            scale (float): Fraction of the width which you wish
                the figure to occupy
            ratio (float): height to width ratio
            post_process (bool): switch post-processing of pgfs on and of
            substitutes (list of tuple of str):
            scaling (str): Default is \\igraph@width
            - further are passed to savefig

    """
    # catch figure from axes or fig
    axes = kwargs.get("axes", None)
    if axes is None:
        fig_default = plt.gcf()
        axes = gca()
        if fig_default is None:
            raise ValueError("fig_default may not be None")
    else:
        fig_default = axes.figure
    fig = kwargs.get("fig", fig_default)

    # resize the figure to match a given width (in pt).
    width = kwargs.pop("width", None)
    scale = kwargs.pop("scale", 1.0)
    orig_size = fig.get_size_inches()
    orig_ratio = orig_size[1] / orig_size[0]
    ratio = kwargs.pop("ratio", orig_ratio)
    if width is not None:
        fig.set_size_inches(*rna.plotting.figsize(width, scale=scale, ratio=ratio))

    # crop the plot down based on the extents of the artists in the plot
    kwargs["bbox_inches"] = kwargs.pop("bbox_inches", "tight")
    if kwargs["bbox_inches"] == "tight":
        extra_artists = None
        for ax in fig.get_axes():
            first_label = ax.get_legend_handles_labels()[0] or None
            if first_label:
                if not extra_artists:
                    extra_artists = []
                if isinstance(first_label, list):
                    extra_artists.extend(first_label)
                else:
                    extra_artists.append(first_label)
        kwargs["bbox_extra_artists"] = kwargs.pop("bbox_extra_artists", extra_artists)

    # pgf post-processing works but matpltolib pgf backend writes pgftext left
    # adjusted and creates right adjusted text by moving the text in x
    # this makes it impossible to rescale text correctly. If one knew, the
    # original was right adjusted a correction could be applied. This way
    # however, we cann not do anything.
    post_process = kwargs.pop("post_process", False)
    pgf_kwargs = {
        "substitutes": kwargs.pop("substitutes", [(r"\sffamily", "")]),
        "scaling": kwargs.pop("scaling", r"\igraph@width"),
    }

    # set current figure
    plt.figure(fig.number)

    if path.endswith("tikz"):
        import matplotlib2tikz as tikz

        if width is None:
            figurewidth = r"\textwidth"
        else:
            figurewidth = "{0}pt".format(scale * width)
        tikz.save(path, figurewidth=figurewidth)
    else:
        plt.savefig(path, **kwargs)

    if post_process and path.endswith("pgf"):
        # use it like \igraph[width=0.5\columnwidth]{path_to_the.pgf}
        substitutes = pgf_kwargs.pop("substitutes")
        scaling = pgf_kwargs.pop("scaling")
        with open(path, "r") as f:
            content = []
            exp = r"(\\pgftext.+)(\\fontsize{.+)\\selectfont(.+)}%"
            exp_width = (
                r"\\pgfpathrectangle{\\pgfpointorigin}" r"{\\pgfqpoint{(.+)}{(.+)}%"
            )
            for line in f:
                res = re.match(exp, line)
                res_width = re.match(exp_width, line)
                if res_width:
                    pgf_path_size = res_width.groups()[0]
                    # factor 72.27 (latex definition differs from pdf
                    # definition) for conversion from in to pt
                    pgf_path_size = r"(" + pgf_path_size.replace("in", "*72.27") + r")"
                if res:
                    groups = res.groups()
                    line = (
                        groups[0]
                        + r"\selectfont \scalebox{\fpeval{(%s/(%s))}}{"
                        % (pgf_path_size, scaling)
                        + groups[2]
                        + r"}}%"
                        + "\n"
                    )
                for x, y in substitutes:
                    line = line.replace(x, y)
                content.append(line)
        with open(path, "w") as f:
            f.writelines(content)

    # reset the figure size
    if width is not None:
        fig.set_size_inches(*orig_size)


def plot_array(array, **kwargs):
    """
    Points3D plotting method.

    Args:
        array (numpy array)
        axes (matplotlib.Axis) object
        x_index (int): coordinate index that should be on x_index
        y_index (int): coordinate index that should be on y_index
        z_index (int or None): coordinate index that should be on z_index.
            If it evaluates to None, 2D plot will be done.
        method (str | callable): method name to use for filling the axes

    Returns:
        Artist or list of Artists (imitating the axes.scatter/plot behaviour).
        Better Artist not list of Artists
    """
    array = np.array(array)
    kwargs.setdefault("method", "scatter")
    plot_manager = rna.plotting.PlotManager(kwargs)

    labels = plot_manager.pop("labels", ["x (m)", "y (m)", "z (m)"])
    x_index, y_index, z_index = plot_manager.get_axis_indices()
    set_labels(plot_manager.axes, *plot_manager.sort_labels(labels))
    if z_index is None:
        args = [array[:, x_index], array[:, y_index]]
    else:
        args = [array[:, x_index], array[:, y_index], array[:, z_index]]
    artist = plot_manager.method(*args, **plot_manager.plot_kwargs)
    return artist


def plot_mesh(vertices, faces, **kwargs):
    """
    Args:
        axes (matplotlib axes)
        x_index (int)
        y_index (int)
        z_index (int)
        edgecolor (color)
        color (color): if given, use this color for faces in 2D
        cmap
        vmin
        vmax
    """
    vertices = np.array(vertices)
    faces = np.array(faces)
    if faces.shape[0] == 0:
        warnings.warn("No faces to plot")
        return None
    if max(faces.flat) > vertices.shape[0]:
        raise ValueError("Some faces point to non existing vertices.")
    po = rna.plotting.PlotManager(kwargs)
    if po.dim is None:
        po.dim = 2
    if po.dim == 2:
        full = True
        import tfields

        mesh = tfields.Mesh3D(vertices, faces=faces)
        x_index, y_index, z_index = po.get_axis_indices()
        facecolors = po.retrieve_chain("facecolors", "color", default=0, keep=False)
        if full:
            # implementation that will sort the triangles by z_index
            centroids = mesh.centroids()
            axis_indices = [0, 1, 2]
            axis_indices.pop(axis_indices.index(x_index))
            axis_indices.pop(axis_indices.index(y_index))
            z_index = axis_indices[0]
            zs = centroids[:, z_index]
            try:
                # no numbers
                iter(facecolors)
                # no strings
                if isinstance(facecolors, str):
                    raise TypeError()
            except TypeError:
                zs, faces = tfields.lib.util.multi_sort(zs, faces)
            else:
                zs, faces, facecolors = tfields.lib.util.multi_sort(
                    zs, faces, facecolors
                )

            n_faces_initial = len(faces)
        else:
            # cut away "back sides" implementation
            direction_vector = np.array([1.0, 1.0, 1.0])
            direction_vector[x_index] = 0.0
            direction_vector[y_index] = 0.0
            norm_vectors = mesh.triangles().norms()
            dot_product = np.dot(norm_vectors, direction_vector)
            n_faces_initial = len(faces)
            faces = faces[dot_product > 0]

        vertices = mesh

        po.plot_kwargs["method"] = "tripcolor"
        po.plot_kwargs["triangles"] = faces

        """
        sort out color arguments
        """
        facecolors = po.format_colors(facecolors, fmt="norm", length=n_faces_initial)
        if not full:
            facecolors = facecolors[dot_product > 0]
        po.plot_kwargs["facecolors"] = facecolors

        d = po.plot_kwargs
        d["x_index"] = x_index
        d["y_index"] = y_index
        artist = plot_array(vertices, **d)
    elif po.dim == 3:
        label = po.pop("label", None)
        color = po.retrieve_chain(
            "color", "c", "facecolors", default="grey", keep=False
        )
        from rna.plotting.colors import color_fmt

        orig_fmt = color_fmt(color)
        orig_color = color

        nanMask = np.isnan(color)
        if nanMask.any():
            warnings.warn("nan found in colors. Removing corresponding faces!")
            color = color[~nanMask]
            faces = faces[~nanMask]

        color = po.format_colors(color, fmt="rgba", length=len(faces))

        edgecolor = po.pop("edgecolor", None)
        alpha = po.pop("alpha", None)
        # delete norm arguments
        po.pop("vmin", None)
        po.pop("vmax", None)
        po.pop("cmap", None)

        triangles = np.array([vertices[face] for face in faces])
        artist = plt3D.art3d.Poly3DCollection(triangles, **po.plot_kwargs)
        po.axes.add_collection3d(artist)

        if edgecolor is not None:
            artist.set_edgecolor(edgecolor)
            artist.set_facecolors(color.tolist())
        else:
            artist.set_color(color.tolist())
        if orig_fmt == "norm":
            # sets vmin, vmax to min, max of val
            artist.set_array(np.array(orig_color))
        if alpha is not None:
            artist.set_alpha(alpha)

        # for some reason auto-scale does not work
        autoscale_3d(po.axes, array=vertices)

        # legend lables do not work at all as an argument
        if label:
            artist.set_label(label)

        # when plotting the legend edgecolors/facecolors2d are needed
        artist._edgecolors2d = None
        artist._facecolors2d = None

        labels = ["x (m)", "y (m)", "z (m)"]
        set_labels(po.axes, *po.sort_labels(labels))
    else:
        raise NotImplementedError("Dimension != 2|3")

    return artist


def plot_tensor_field(points, field, **kwargs):
    """
    Args:
        points (array_like): base vectors
        field (): direction field
        **kwargs: passed to PlotManager
    """
    po = rna.plotting.PlotManager(kwargs)
    field = np.array(field)
    if len(field.shape) == 2 and field.shape[1] == 1:
        # scalar
        field = field.reshape(len(points))
    if len(field.shape) == 1:
        # scalar
        colors = po.format_colors(field)
        cmap, vmin, vmax = po.pop_norm_args()
        artists = plot_array(points, color=colors, **po.state)
        if vmin is not None:
            field[np.argwhere(field < vmin)] = vmin
        if vmax is not None:
            field[np.argwhere(field > vmax)] = vmax
        if not isinstance(artists, list):
            artists.set_array(field)
        else:
            raise NotImplementedError(
                "Probably you used plot instead of"
                "scatter. This is not yet implemented because you have n-1"
                "segments but gave n colors."
            )
    elif len(field.shape) == 2:
        # vector
        if points is None:
            points = np.full(field.shape, 0.0)
        x_index, y_index, z_index = po.get_axis_indices()
        artists = []
        for point, vector in zip(points, field):
            if po.dim == 3:
                artists.append(
                    po.axes.quiver(
                        point[x_index],
                        point[y_index],
                        point[z_index],
                        vector[x_index],
                        vector[y_index],
                        vector[z_index],
                        **po.plot_kwargs
                    )
                )
            elif po.dim == 2:
                artists.append(
                    po.axes.quiver(
                        point[x_index],
                        point[y_index],
                        vector[x_index],
                        vector[y_index],
                        **po.plot_kwargs
                    )
                )
            else:
                raise NotImplementedError("Dimension != 2|3")
    else:
        raise NotImplementedError("Only Scalars and Vectors implemented")
    return artists


def plot_plane(point, normal, **kwargs):
    def plot_vector(fig, orig, v, color="blue"):
        axes = fig.gca(projection="3d")
        orig = np.array(orig)
        v = np.array(v)
        axes.quiver(orig[0], orig[1], orig[2], v[0], v[1], v[2], color=color)
        axes.set_xlim(0, 10)
        axes.set_ylim(0, 10)
        axes.set_zlim(0, 10)
        axes = fig.gca(projection="3d")
        return fig

    def rotation_matrix(d):
        sin_angle = np.linalg.norm(d)
        if sin_angle == 0:
            return np.identity(3)
        d /= sin_angle
        eye = np.eye(3)
        ddt = np.outer(d, d)
        skew = np.array(
            [[0, d[2], -d[1]], [-d[2], 0, d[0]], [d[1], -d[0], 0]], dtype=np.float64
        )

        M = ddt + np.sqrt(1 - sin_angle ** 2) * (eye - ddt) + sin_angle * skew
        return M

    def pathpatch_2d_to_3d(pathpatch, z, normal):
        if type(normal) is str:  # Translate strings to normal vectors
            index = "xyz".index(normal)
            normal = np.roll((1.0, 0, 0), index)

        normal /= np.linalg.norm(normal)  # Make sure the vector is normalised
        path = pathpatch.get_path()  # Get the path and associated transform
        trans = pathpatch.get_patch_transform()

        path = trans.transform_path(path)  # Apply the transform

        pathpatch.__class__ = plt3D.art3d.PathPatch3D  # Change the class
        pathpatch._code3d = path.codes  # Copy the codes
        pathpatch._facecolor3d = pathpatch.get_facecolor  # Get the face color

        verts = path.vertices  # Get the vertices in 2D

        d = np.cross(normal, (0, 0, 1))  # Obtain the rotation vector
        M = rotation_matrix(d)  # Get the rotation matrix

        pathpatch._segment3d = np.array(
            [np.dot(M, (x, y, 0)) + (0, 0, z) for x, y in verts]
        )

    def pathpatch_translate(pathpatch, delta):
        pathpatch._segment3d += delta

    kwargs["alpha"] = kwargs.pop("alpha", 0.5)
    po = rna.plotting.PlotManager(kwargs)
    patch = Circle((0, 0), **po.plot_kwargs)
    po.axes.add_patch(patch)
    pathpatch_2d_to_3d(patch, z=0, normal=normal)
    pathpatch_translate(patch, (point[0], point[1], point[2]))


def plot_sphere(point, radius, **kwargs):
    po = rna.plotting.PlotManager(kwargs)
    # Make data
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)
    x = point[0] + radius * np.outer(np.cos(u), np.sin(v))
    y = point[1] + radius * np.outer(np.sin(u), np.sin(v))
    z = point[2] + radius * np.outer(np.ones(np.size(u)), np.cos(v))

    # Plot the surface
    return po.axes.plot_surface(x, y, z, **po.plot_kwargs)


def plot_function(fun, **kwargs):
    """
    Args:
        axes (matplotlib.Axis) object

    Returns:
        Artist or list of Artists (imitating the axes.scatter/plot behaviour).
        Better Artist not list of Artists
    """
    import numpy as np

    labels = ["x", "f(x)"]
    po = rna.plotting.PlotManager(kwargs)
    rna.plotting.set_labels(po.axes, *labels)
    xMin, xMax = po.pop("xMin", 0), po.pop("xMax", 1)
    n = po.pop("n", 100)
    vals = np.linspace(xMin, xMax, n)
    args = (vals, map(fun, vals))
    artist = po.axes.plot(*args, **po.plot_kwargs)
    return artist


def plot_errorbar(points, errors_up, errors_down=None, **kwargs):
    """
    Args:
        axes (matplotlib.Axis) object

    Returns:
        Artist or list of Artists (imitating the axes.scatter/plot behaviour).
        Better Artist not list of Artists
    """
    po = rna.plotting.PlotManager(kwargs)
    po.setdefault("marker", "_")

    if errors_down is None:
        errors_down = errors_up

    artists = []

    # plot points
    # artists.append(po.axes.plot(points, **po.plot_kwargs))

    # plot errorbars
    for i in range(len(points)):
        artists.append(
            po.axes.plot(
                [points[i, 0] + errors_up[i, 0], points[i, 0] - errors_down[i, 0]],
                [points[i, 1], points[i, 1]],
                [points[i, 2], points[i, 2]],
                **po.plot_kwargs
            )
        )
        artists.append(
            po.axes.plot(
                [points[i, 0], points[i, 0]],
                [points[i, 1] + errors_up[i, 1], points[i, 1] - errors_down[i, 1]],
                [points[i, 2], points[i, 2]],
                **po.plot_kwargs
            )
        )
        artists.append(
            po.axes.plot(
                [points[i, 0], points[i, 0]],
                [points[i, 1], points[i, 1]],
                [points[i, 2] + errors_up[i, 2], points[i, 2] - errors_down[i, 2]],
                **po.plot_kwargs
            )
        )

    return artists


"""
Display section
"""


def axes_dim(axes):
    """
    Returns int: axes dimension
    """
    if hasattr(axes, "get_zlim"):
        return 3
    else:
        return 2


def set_aspect_equal(axes):
    """Fix equal aspect bug for 3D plots."""

    if axes_dim(axes) == 2:
        axes.set_aspect("equal")
        return

    xlim = axes.get_xlim3d()
    ylim = axes.get_ylim3d()
    zlim = axes.get_zlim3d()

    from numpy import mean

    xmean = mean(xlim)
    ymean = mean(ylim)
    zmean = mean(zlim)

    plot_radius = max(
        [
            abs(lim - mean_)
            for lims, mean_ in ((xlim, xmean), (ylim, ymean), (zlim, zmean))
            for lim in lims
        ]
    )

    axes.set_xlim3d([xmean - plot_radius, xmean + plot_radius])
    axes.set_ylim3d([ymean - plot_radius, ymean + plot_radius])
    axes.set_zlim3d([zmean - plot_radius, zmean + plot_radius])


def set_axis_off(axes):
    if axes_dim(axes) == 2:
        axes.set_axis_off()
    else:
        axes._axis3don = False


def autoscale_3d(axes, array=None, xLim=None, yLim=None, zLim=None):
    if array is not None:
        xMin, yMin, zMin = array.min(axis=0)
        xMax, yMax, zMax = array.max(axis=0)
        xLim = (xMin, xMax)
        yLim = (yMin, yMax)
        zLim = (zMin, zMax)
    xLimAxis = axes.get_xlim()
    yLimAxis = axes.get_ylim()
    zLimAxis = axes.get_zlim()

    if not False:
        # not empty axes
        xMin = min(xLimAxis[0], xLim[0])
        yMin = min(yLimAxis[0], yLim[0])
        zMin = min(zLimAxis[0], zLim[0])
        xMax = max(xLimAxis[1], xLim[1])
        yMax = max(yLimAxis[1], yLim[1])
        zMax = max(zLimAxis[1], zLim[1])
    axes.set_xlim([xMin, xMax])
    axes.set_ylim([yMin, yMax])
    axes.set_zlim([zMin, zMax])


def set_legend(axes, artists, **kwargs):
    """
    Convenience method to set a legend from multiple artists to an axes.
    Args:
        **kwargs
            table (bool): if True, labels containing ',' are mapped to table
            table_title (str): value of the table entry top left - only active
                if table
    Examples:
        >> import rna
        >> import matplotlib.pyplot as plt
        >> import pylab

        >> fig = plt.figure()
        >> ax = fig.add_subplot(111)

        >> im1 = ax.plot(range(10), pylab.randn(10), "r--",
        ...              label=(r"$i = 1$,$j = 1$"))
        >> im2 = ax.plot(range(10), pylab.randn(10), "g--",
        ...              label=(r"$i = 1$,$j = 2$"))
        >> im3 = ax.plot(range(10), pylab.randn(10), "b--",
        ...              label=(r"$i = 1$,$j = 3$"))
        >> im4 = ax.plot(range(10), pylab.randn(10), "r.",
        ...              label=(r"$i = 2$,$j = 1$"))
        >> im5 = ax.plot(range(10), pylab.randn(10), "g.",
        ...              label=(r"$i = 2$,$j = 2$"))
        >> im6 = ax.plot(range(10), pylab.randn(10), "b.",
        ...              label=(r"$i = 2$,$j = 3$"))
        >> im7 = ax.plot(range(10), pylab.randn(10), "r^",
        ...              label=(r"$i = 3$,$j = 1$"))
        >> im8 = ax.plot(range(10), pylab.randn(10), "g^",
        ...              label=(r"$i = 3$,$j = 2$"))
        >> im9 = ax.plot(range(10), pylab.randn(10), "b^",
        ...              label=(r"$i = 3$,$j = 3$"))
        >> handles = [im1, im2, im3, im4, im5, im6, im7, im8, im9]

        >> rna.plotting.set_legend(ax, handles, table=True)

        >> plt.show()
    """
    table = kwargs.pop("table", False)
    labels = kwargs.pop("labels", None)
    ncol = kwargs.pop("ncol", None)

    handles = []
    for artist in artists:
        if isinstance(artist, list):
            handles.append(artist[0])
        elif isinstance(artist, tuple):
            tuple_handle = tuple()
            for sub_artist in artist:
                if isinstance(sub_artist, list):
                    tuple_handle += (sub_artist[0],)
                else:
                    tuple_handle += (sub_artist,)
            handles.append(tuple_handle)
        else:
            handles.append(artist)

    if labels is None and any([isinstance(h, tuple) for h in handles]):
        labels = []
        for h in handles:
            if isinstance(h, tuple):
                sub_labels = [sub_h.get_label() for sub_h in h]
                label = sub_labels[0]
                for sub_label in sub_labels[1:]:
                    if not sub_label.startswith("_"):
                        label += sub_label
            else:
                label = h.get_label()
            labels.append(label)

    if table and labels is None and ncol is None:
        table_title = kwargs.pop("table_title", "")
        labels = np.array([h.get_label() for h in handles])
        labels = [label.split(",") for label in labels]
        captions_i = []
        captions_j = []
        for label in labels:
            if label[0] not in captions_i:
                captions_i.append(label[0])
            if label[1] not in captions_j:
                captions_j.append(label[1])

        shape = (len(captions_i), len(captions_j))

        # initialize
        shape = np.array(shape)
        handles = np.array(handles)

        # create blank rectangle
        extra = Rectangle(
            (0, 0), 1, 1, fc="w", fill=False, edgecolor="none", linewidth=0
        )

        # Create organized list containing all handles for table.
        # Extra represent empty space
        handles_table = np.full(shape + 1, extra, dtype=object)
        for handle, label in zip(handles, labels):
            i = captions_i.index(label[0])
            j = captions_j.index(label[1])
            if handles_table[i + 1, j + 1] != extra:
                raise ValueError("Duplicate label {label}".format(**locals()))
            handles_table[i + 1, j + 1] = handle

        # Define the label captions
        labels_table = np.full(shape + 1, "", dtype="S80")
        labels_table[0, 0] = table_title
        labels_table[0, 1:] = captions_j
        labels_table[1:, 0] = captions_i
        labels_table = labels_table.astype(str)

        handles = list(handles_table.flat)
        labels = list(labels_table.flat)
        kwargs["ncol"] = shape[0] + 1
        # negative numbers in handletextpad move to the right
        kwargs["handletextpad"] = kwargs.pop("handletextpad", -1.5)
        kwargs["columnspacing"] = kwargs.pop("columnspacing", 1.5)

    return axes.legend(handles=handles, labels=labels, **kwargs)


class set_zoomable(object):
    """
    Left click a colorbar and release in order to zoom.
    Upper and lower 5% of the colorbar will zoom out.
    """

    def __init__(self, cbar):
        self._cbar = cbar
        self._v_press = None
        self._v_release = None

        artist = self._cbar.mappable
        self._press_connection_id = artist.axes.figure.canvas.mpl_connect(
            "button_press_event", partial(self.on_press)
        )
        self._release_connection_id = artist.axes.figure.canvas.mpl_connect(
            "button_release_event", partial(self.on_release)
        )

    def on_press(self, event):
        if event.inaxes is self._cbar.ax:
            self._v_press = event.ydata

    def on_release(self, event):
        if event.inaxes is self._cbar.ax:
            self._v_release = event.ydata
            self.update_v_min_max()

    def update_v_min_max(self):
        # sort press and release event
        if self._v_press > self._v_release:
            x_up = self._v_press
            x_low = self._v_release
        if self._v_press < self._v_release:
            x_up = self._v_release
            x_low = self._v_press

        # zoom out if in 5% margin
        if x_up > 0.95:
            x_up = 1.5
        if x_low < 0.05:
            x_low = -0.5

        artist = self._cbar.mappable
        vmin, vmax = artist.get_clim()
        v_range = vmax - vmin
        vmax = vmin + x_up * v_range
        vmin = vmin + x_low * v_range
        artist.set_clim(vmin, vmax)
        artist.axes.figure.canvas.draw()


def set_colorbar(
    axes,
    artist,
    label=None,
    divide=True,
    position="right",
    size="2%",
    pad=0.05,
    labelpad=None,
    zoom=False,
    **kwargs
):
    """
    Note:
        Bug found in matplotlib:
            when calling axes.clear(), the colorbar has to be removed by hand,
            since it will not be removed by clear.
        >> import rna
        >> axes = rna.plotting.gca()
        >> artist = ...
        >> cbar = rna.plotting.set_colorbar(axes, artist)
        >> rna.plotting.save(...)
        >> cbar.remove()  # THIS IS IMPORTANT. Otherwise you will have problems
        # at the next call of savefig.
        >> axes.clear()

    """
    ticks_position = "default"
    label_position = "bottom"
    labelpad = 30 if labelpad is None else labelpad
    if position == "right":
        rotation = 270
    elif position == "left":
        rotation = 90
    elif position == "top":
        rotation = 0
        ticks_position = "top"
        label_position = "top"
        labelpad = 5
    elif position == "bottom":
        rotation = 0
    # colorbar
    if divide and axes_dim(axes) == 2:
        divider = make_axes_locatable(axes)
        axes = divider.append_axes(position, size=size, pad=pad)
        cbar = plt.colorbar(artist, cax=axes, **kwargs)
    else:
        cbar = plt.colorbar(artist, ax=axes, **kwargs)
    cbar.ax.xaxis.set_ticks_position(ticks_position)
    cbar.ax.xaxis.set_label_position(label_position)
    cbar.ax.tick_params(axis="x", which="major", pad=0)

    # label
    if label is None:
        art_label = artist.get_label()
        if art_label:
            label = art_label
    if label is not None:
        cbar.set_label(label, rotation=rotation, labelpad=labelpad)

    if zoom:
        set_zoomable(cbar)

    return cbar


def set_labels(axes, *labels):
    """
    Set x, y and possibly z labels
    """
    axes.set_xlabel(labels[0])
    axes.set_ylabel(labels[1])
    if axes_dim(axes) == 3:
        axes.set_zlabel(labels[2])
    elif axes_dim(axes) != 2:
        raise ValueError(
            "Axis dimensions other than 2 or 3 is not supported" "for set_labels."
        )


def set_formatter(sub_axes=None, formatter=dates.DateFormatter("%d-%m-%y")):
    if sub_axes is None:
        axes = gca()
        sub_axes = axes.xaxis
    sub_axes.set_major_formatter(formatter)


class ScientificFormatter(mpl.ticker.ScalarFormatter):
    """
    Examples:
        >> cbar = rna.plotting.set_colorbar(
        ...     axes, artist,
        ...     label=r"$q_c\;(MW/m^2)$",  # NOQA
        ...     format=rna.plotting.backend.ScientificFormatter(
        ...         None, useMathText=False))

    Args:
        oom (int): order of magnitued on the axes
        **kwargs: forwarded to ScalarFormatter

    """

    def __init__(self, oom=None, **kwargs):
        self._oom = oom
        super(ScientificFormatter, self).__init__(**kwargs)

    def _set_orderOfMagnitude(self, oom):
        self._exp = int(np.log10(oom))
        if self._oom is not None:
            oom = self._oom
        else:  # Default: -3, 0, 3, 6, ...
            oom = self._exp - (self._exp % 3)
        self.orderOfMagnitude = oom


# ANIMATION

try:
    FFMPEGWRITER = manimation.writers["ffmpeg"]
    FFMPEG = True
except Exception as err:
    logging.getLogger().warning(str(err))
    FFMPEGWRITER = object
    FFMPEG = False


class Writer(FFMPEGWRITER):
    """
    FFMPEGWRITER child

    Args:
        **kwargs:
            title (str)
            comment (str)
            fps (int)

    Examples:
        >>> import numpy as np
        >>> import rna
        >>> axes = rna.plotting.gca()
        >>> writer = rna.plotting._backend_mod.Writer()
        >>> pictures = np.array([[[1,1], [0,0]],
        ...                      [[0.5,1], [0,0.5]]])
        >>> with writer.saving():
        ...     for s in pictures:
        ...         artist = axes.imshow(s.T)
        ...         writer.grab_frame()
        ...         axes.clear()
    """

    def __init__(self, **kwargs):
        if FFMPEG:
            # set Defaults
            title = kwargs.pop("title", "")
            comment = kwargs.pop("comment", "")
            artist = kwargs.pop("artist", "rna plotting module")
            fps = kwargs.pop("fps", 15)
            if len(kwargs) > 0:
                raise ValueError("Some Attributes are not used.")

            # set up writer
            metadata = dict(title=title, artist=artist, comment=comment)
            super(Writer, self).__init__(fps=fps, metadata=metadata)

    @contextmanager
    def saving(self, fig=None, path=None, dpi=300, **kwargs):
        """
        Overwritinig to FFMPEGWRITER.saving
        setting defaults though
        """
        log = logging.getLogger()
        if FFMPEG:
            if fig is None:
                fig = plt.gcf()
            if path is None:
                path = "/tmp/lastMovie.mp4"
            else:
                path = rna.path.resolve(path)
            log.info("Write movie to {path}".format(**locals()))
            with super(Writer, self).saving(fig, path, dpi, **kwargs):
                yield
        else:
            log.warn("FFMPEG Writer not available. Nothing is written!")


if __name__ == "__main__":
    import tfields

    m = tfields.Mesh3D.grid((0, 2, 2), (0, 1, 3), (0, 0, 1))
    m.maps[0].fields.append(tfields.Tensors(np.arange(m.faces.shape[0])))
    art1 = m.plot(dim=3, map_index=0, label="twenty")

    m = tfields.Mesh3D.grid((4, 7, 2), (3, 5, 3), (2, 2, 1))
    m.maps[0].fields.append(tfields.Tensors(np.arange(m.faces.shape[0])))
    art = m.plot(dim=3, map_index=0, edgecolor="k", vmin=-1, vmax=1, label="something")

    plot_sphere([7, 0, 1], 3)



"""
Core plotting tools for tfields library. Especially PlotOptions class
is basis for many plotting expansions

TODO:
    * add other library backends. Do not restrict to mpl
"""
import warnings
import matplotlib.pyplot as plt


class PlotManager(object):
    """
    processing kwargs for plotting functions and providing easy
    access to axes, dimension and plotting method as well as indices
    for array choice (x..., y..., z_index)

    Examples:
        >>> import rna

        Dimension arguemnt (dim)
        >>> rna.plotting.use('matplotlib')  # which is default anyway
        >>> pm = rna.plotting.PlotManager({})
        >>> pm.dim
        2
        >>> rna.plotting.axes_dim(pm.axes)
        2

        >>> pm = rna.plotting.PlotManager({'dim': 3})
        >>> pm.dim
        3
        >>> rna.plotting.axes_dim(pm.axes)
        3

        >>> rna.plotting.use('pyqtgraph')
        >>> pm = rna.plotting.PlotManager({})
        >>> pm.dim
        3
        >>> rna.plotting.axes_dim(pm.axes)
        3

        Passing an axes object specifically
        >>> rna.plotting.use('matplotlib')
        >>> axes = rna.plotting.gca(3)
        >>> pm = rna.plotting.PlotManager({'axes': axes})

    """
    axes_default = None
    dim_default = None
    x_index_default = 0
    y_index_default = 1
    z_index_default = None
    method_default = None

    def __init__(self, kwargs):
        kwargs = dict(kwargs)

        # axes are generated with priority kwargs 'axes', 'dim', 'z_index'
        self._axes = self.axes_default
        self._dim = self.dim_default
        self._x_index = kwargs.pop('x_index', self.x_index_default)
        self._y_index = kwargs.pop('y_index', self.y_index_default)
        self._z_index = kwargs.pop('z_index', self.z_index_default)
        if 'axes' in kwargs:
            self.axes = kwargs.pop('axes')
            dim = kwargs.pop('dim', self.dim)
            if dim != self.dim:
                raise ValueError("Axes dimension and dim argument in conflict:"
                                 "{self.dim} vs {dim}".format(**locals()))
        elif 'dim' in kwargs:
            self.dim = kwargs.pop('dim')
        else:
            import rna.plotting
            self.axes = rna.plotting.gca()
        self.method = kwargs.pop('method', self.method_default)

        self.plot_kwargs = kwargs

    @property
    def method(self):
        """
        Method for plotting. Will be callable together with plot_kwargs
        """
        return self._method

    @method.setter
    def method(self, method):
        if not isinstance(method, str):
            self._method = method
        else:
            self._method = getattr(self.axes, method)

    @property
    def dim(self):
        """
        axes dimension. If not axes object is given, the dimension defaults to
        2 unless z_index is given which means dim=3
        Examples:
            >>> import rna

            >>> kwargs = dict(dim=3)
            >>> pm = rna.plotting.PlotManager(kwargs)
            >>> rna.plotting.axes_dim(pm.axes)
            3

            >>> kwargs = dict(z_index=42)
            >>> pm = rna.plotting.PlotManager(kwargs)
            >>> rna.plotting.axes_dim(pm.axes)
            3

        """
        return self._dim

    @dim.setter
    def dim(self, dim):
        if dim != self._dim:
            # explicit call to change dimension and thus also change axes
            # do not raise warning
            self._dim = dim
            self._z_index = None
            import rna.plotting
            axes = rna.plotting.gca(dim)
            self.axes = axes

        # check and set z_index
        if dim == 2:
            if self._z_index is not None:
                warnings.warn("Dimension is set to 2 but z_index is not None")
            self._z_index = None
        elif dim == 3:
            if self._z_index is None:
                # auto compute the z_index if axes dim is 3
                indices_used = [0, 1, 2]
                indices_used.remove(self._x_index)
                indices_used.remove(self._y_index)
                z_index = indices_used[0]
                self._z_index = z_index
        else:
            raise NotImplementedError("Dimensions other than 2 or 3 are not"
                                      "supported.")
        self._dim = dim

    @property
    def axes(self):
        """
        The Axes object that belongs to this instance
        """
        return self._axes

    @axes.setter
    def axes(self, axes):
        self._axes = axes

        # correct dimensions if necessary
        from rna.plotting import axes_dim
        dim = axes_dim(self._axes)
        if self.dim is not None and self.dim != dim:
            warnings.warn("Axes force dimensions that were set differently"
                          " before.")
        self.dim = dim

    @property
    def state(self):
        """
        Returns:
            dict: dict which can be used to reconstruct the state of the
                current object

        Examples:
            Setting up a PlotManager
            >>> import rna
            >>> ax_tmp = rna.plotting.gca(3)
            >>> kwargs = {'method': 'plot', 'dim': 2}
            >>> plot_manager = rna.plotting.PlotManager(kwargs)
            >>> _ = kwargs.setdefault('method', 'scatter')

            Pass the state arguments to the next PlotManager

            >>> state = plot_manager.state
            >>> plot_manager_2 = rna.plotting.PlotManager(state)
            >>> plot_manager_2.state['method']  # doctest: +ELLIPSIS
            <...plot of <matplotlib.axes._subplots.AxesSubplot object ...>>
            >>> plot_manager_2.state['axes']  # doctest: +ELLIPSIS
            <matplotlib.axes._subplots.AxesSubplot object ...>

            If some argument is None it will be not occuring in the state

            >>> plot_manager.method = None
            >>> assert 'method' not in plot_manager.state

        """
        state_dict = dict(self.plot_kwargs)
        for state_key in ['axes',
                          'x_index', 'y_index', 'z_index',
                          'dim',
                          'method']:
            state_value = getattr(self, '_' + state_key)
            state_value_default = getattr(self, state_key + '_default')
            if state_value != state_value_default:
                state_dict[state_key] = state_value
        return state_dict

    def get_axis_indices(self):
        return self._x_index, self._y_index, self._z_index

    def set_vmin_vmax_auto(self, vmin, vmax, scalars):
        """
        Automatically set vmin and vmax as min/max of scalars
        but only if vmin or vmax is None
        """
        if scalars is None:
            return
        if len(scalars) < 2:
            warnings.warn("Need at least two scalars to autoset vmin and/or"
                          " vmax!")
            return
        if vmin is None:
            vmin = min(scalars)
            self.plot_kwargs['vmin'] = vmin
        if vmax is None:
            vmax = max(scalars)
            self.plot_kwargs['vmax'] = vmax

    def get_norm_args(self, vmin_default=0, vmax_default=1, cmap_default=None):
        """
        Examples:
            >>> import rna
            >>> rna.plotting.set_style()
            >>> pm = rna.plotting.PlotManager(dict(vmin=2, vmax=42))
            >>> pm.get_norm_args()
            ('viridis', 2, 42)
        """
        if cmap_default is None:
            cmap_default = plt.rcParams['image.cmap']
        cmap = self.get('cmap', cmap_default)
        vmin = self.get('vmin', vmin_default)
        vmax = self.get('vmax', vmax_default)
        if vmin is None:
            vmin = vmin_default
        if vmax is None:
            vmax = vmax_default
        return cmap, vmin, vmax

    def pop_norm_args(self, **defaults):
        """
        Pop vmin, vmax and cmap from plot_kwargs
        Args:
            **defaults:
                see get_norm_args method
        """
        cmap, vmin, vmax = self.get_norm_args(**defaults)
        self.pop('cmap')
        self.pop('vmin')
        self.pop('vmax')
        return cmap, vmin, vmax

    def format_colors(self, colors, fmt='rgba', length=None):
        """
        format colors according to fmt argument
        Args:
            colors (list/one value of rgba tuples/int/float/str): This argument
                will be interpreted as color
            fmt (str): rgba | hex | norm
            length (int/None): if not None: correct colors lenght

        Returns:
            colors in fmt
        """
        cmap, vmin, vmax = self.get_norm_args(cmap_default='viridis',
                                              vmin_default=None,
                                              vmax_default=None)
        from rna.plotting.colors import to_colors
        return to_colors(colors, fmt, length=length,
                         vmin=vmin, vmax=vmax, cmap=cmap)

    def sort_labels(self, labels):
        """
        Returns the labels corresponding to the axis_indices
        """
        return [labels[i] for i in self.get_axis_indices() if i is not None]

    def get(self, attr, default=None):
        return self.plot_kwargs.get(attr, default)

    def pop(self, attr, default=None):
        return self.plot_kwargs.pop(attr, default)

    def set(self, attr, value):
        self.plot_kwargs[attr] = value

    def setdefault(self, attr, value):
        self.plot_kwargs.setdefault(attr, value)

    def retrieve(self, attr, default=None, keep=True):
        if keep:
            return self.get(attr, default)
        else:
            return self.pop(attr, default)

    def retrieve_chain(self, *args, **kwargs):
        default = kwargs.pop('default', None)
        keep = kwargs.pop('keep', True)
        if len(args) > 1:
            return self.retrieve(args[0],
                                 self.retrieve_chain(*args[1:],
                                                     default=default,
                                                     keep=keep),
                                 keep=keep)
        if len(args) != 1:
            raise ValueError("Invalid number of args ({0})".format(len(args)))
        return self.retrieve(args[0], default, keep=keep)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
    # doctest.run_docstring_examples(as_tensors_list, globals())



"""Top-level package for rna."""

__author__ = """Daniel Böckenhoff"""
__email__ = "dboe@ipp.mpg.de"
__version__ = "0.5.0"

from . import plotting
from . import log  # NOQA
from . import path  # NOQA

plotting.use("matplotlib")



"""
This module is built to be able to use multiple plotting backends in the same
way.
No matter, which backend you use, plotting should work with the steps shown in
the following
Examples:
    >>> import rna
    >>> import tfields

    Build a cube
    >>> cube = tfields.Mesh3D.grid(
    ...         (-1, 1, 2),
    ...         (-1, 1, 2),
    ...         (-5, -3, 2))
    >>> cube.maps[3].fields = [tfields.Tensors(list(range(len(cube.maps[3]))))]
    >>> map_index = 0
    >>> scalars = cube.maps[3].fields[map_index]

    Tell rna to use the pyqtgraph backend. Default is matplotlib
    >>> rna.plotting.use('pyqtgraph')

    Get an axes element
    >>> axes = rna.plotting.gca(3)

    Plot the cube
    >>> artist = rna.plotting.plot_mesh(cube, cube.maps[3], color=scalars,
    ...     cmap='plasma', axes=axes)

    Plot a colorbar
    >>> cbar = rna.plotting.set_colorbar(axes, artist)

    Show the app and start the loop. Commented for doctests
    # >>> rna.plotting.show()

    Switching back to default
    >>> rna.plotting.use('matplotlib')

"""
import logging
import importlib
import numpy as np
from . import base  # NOQA
from . import colors  # NOQA
from .plot_manager import PlotManager  # NOQA
import rna.path
_log = logging.getLogger(__name__)


backend_methods = [
    'axes_dim',
    'gca',
    'show',
    'clear',
    'plot_mesh',
    'plot_array',
    'plot_tensor_field',
    'set_colorbar',
    '_save',
    'set_style',
    'set_labels',
    'set_aspect_equal',
]
global _backend_mod
_backend_mod = None
for method_name in backend_methods:
    globals()[method_name] = None


def not_implemented_factory(method_name):
    def raising_method(*args, **kwargs):
        raise NotImplementedError(
            "Backend does not provide method '{method_name}'"
            .format(method_name=method_name))
    return raising_method


def use(newbackend):
    """
    Activate a backend of you choice. Non specific methods will be exposed to
    the namespace of rna.plotting . Further methods are exposed to
    rna.plotting.backend
    Args:
        newbackend (str):
            The name of the backend to use.
    Examples:
        >>> import rna
        >>> rna.plotting.use('matplotlib')
        >>> mpl_show = rna.plotting.show
        >>> rna.plotting.use('pyqtgraph')
        >>> pqg_show = rna.plotting.show
        >>> assert callable(mpl_show)
        >>> assert callable(pqg_show)
        >>> assert mpl_show is not pqg_show
        >>> rna.plotting.use('matplotlib')

    """
    backend_name = ("rna.plotting.backend_{}".format(newbackend.lower()))

    backend_mod = importlib.import_module(backend_name)
    import rna.plotting
    # build a new class Backend which gives it module like behaviour
    global backend
    backend = type(
        "Backend", (rna.plotting.base._Backend,), vars(backend_mod))
    _log.debug("Loaded backend %s.", newbackend)

    # expose backend_methods to global namespace of rna.plotting
    global _backend_mod
    _backend_mod = backend_mod
    for method_name in backend_methods:
        if hasattr(backend, method_name):
            # globals()[method_name] = getattr(backend, method_name)
            setattr(rna.plotting,
                    method_name,
                    getattr(_backend_mod, method_name))
        else:
            setattr(rna.plotting,
                    method_name,
                    not_implemented_factory(method_name))
    # Need to keep a global reference to the backend for compatibility reasons.
    # See https://github.com/matplotlib/matplotlib/issues/6092
    rna.plotting.active_backend = newbackend


def save(path, *fmts, **kwargs):
    """
    Args:
        path (str): path without extension to save to
        *fmts (str): format of the figure to save. If multiple are given,
            create that many files
        **kwargs:
            axes
            fig
    """
    log = logging.getLogger()

    if len(fmts) != 0:
        for fmt in fmts:
            if path.endswith('.'):
                new_file_path = path + fmt
            elif '{fmt}' in path:
                new_file_path = path.format(**locals())
            else:
                new_file_path = path + '.' + fmt
            save(new_file_path, **kwargs)
    else:
        path = rna.path.resolve(path)
        log.info("Saving figure as {0}".format(path))
        _save(path, **dict(kwargs))  # copy the dict before passing  # NOQA


def figsize(width_pt, scale=1., ratio=None):
    """
    width_pt (float): figure width width in pt. Can be scaled by
        Get textwidth from latex using '\the\textwidth'
    scale (float)
    ratio (float | None):
        float: ratio of figure height / figure width
        None: golden ratio
    """
    inches_per_pt = 1.0/72.27                       # Convert pt to inch
    if ratio is None:
        # golden ratio
        ratio = (np.sqrt(5.0)-1.0)/2.0
    fig_width = width_pt * inches_per_pt * scale    # fig_width in inches
    fig_height = fig_width * ratio                  # height in inches
    fig_size = (fig_width, fig_height)
    return fig_size


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Input output utilities
"""
import os
import shutil
import functools
import subprocess
import contextlib
import logging

try:
    from pathlib import Path as PathlibPath

    PathlibPath().expanduser()  # not existing in python2 version of pathlib
except (ImportError, AttributeError):
    from pathlib2 import Path as PathlibPath
import glob


class Path(type(PathlibPath())):
    def resolve(self):
        """
        TODO: look at this https://bugs.python.org/issue19776
            There are good reasons that Pathlib doesnt do that
        """
        new_path = self.expanduser()
        return super(Path, new_path).resolve()


def resolve(path, *joins):
    """
    Returns:
        path with resolved ~, symbolic links and relative paths.
        *joins: parts to join with path

    Examples:
        Resolves relative paths
        >>> import os
        >>> from rna.path import resolve
        >>> resolve("../rna") == resolve(".")
        True

        Resolves user variables
        >>> resolve("~") == os.path.expanduser("~")
        True

        Also resolves symlincs which i will not test.
    """
    if joins:
        for part in joins:
            path = os.path.join(path, part)
    # py27 pathlib.Path has no expanduser
    # return str(Path(path).expanduser().resolve())
    # resolve environment variable like ~
    path = os.path.expanduser(str(path))
    # make it absolute
    path = os.path.abspath(path)
    # resolve symlinks but preserve '//' in the beginning
    is_nas_path = path.startswith("//")
    path = os.path.realpath(path)
    if is_nas_path:
        path = "/" + path

    return path


def cp(source, dest, overwrite=True):
    """
    copy with shutil
    """
    source = resolve(source)
    dest = resolve(dest)
    if os.path.isfile(dest):
        if not overwrite:
            raise ValueError(
                "Attempt to overwrite destination path" " {0}.".format(dest)
            )
    if os.path.isfile(source):
        shutil.copy(source, dest)
    elif os.path.isdir(source):
        shutil.copytree(source, dest)
    else:
        raise TypeError("Source {source} is not file or dir".format(**locals()))


def scp(source, dest):
    """
    ssh copy
    """
    # subprocess.check_call does not execute in bash, so it does not know ~
    # Carefull when giving remote host : ~ replacement is not tested
    subprocess.check_call(["scp", source, dest])


def mv(source, dest, overwrite=True):
    """
    Move files or whole folders
    """
    source = resolve(source)
    dest = resolve(dest)
    log = logging.getLogger()
    if os.path.isfile(dest) and not overwrite:
        raise ValueError(
            "Attempting to move to already existing destination"
            " path {0}.".format(dest)
        )
    log.info("Moving file or dir {0} to {1}".format(source, dest))
    if os.path.isfile(source):
        shutil.move(source, dest)
    elif os.path.isdir(source):
        shutil.move(source, dest)
        # shutil.copytree(source, dest)
    else:
        raise TypeError("source is not file or dir")


def rm(path):
    """
    Remove path or (empty) directory
    """
    path = resolve(path)
    if os.path.isfile(path):
        os.remove(path)
    elif os.path.isdir(path):
        if len(ls(path)) > 0:
            raise ValueError("Directory {0} is not empty".format(path))
        os.rmdir(path)
    else:
        raise ValueError("Path {0} is not directory and not file".format(path))


def cd(directory):
    """
    Change directory
    """
    os.chdir(resolve(directory))


@contextlib.contextmanager
def cd_tmp(tmpPath):
    """
    Temporarily change directory. change back afterwards
    """
    cwd = os.getcwd()
    cd(resolve(tmpPath))
    try:
        yield
    finally:
        cd(cwd)


def ls(directory, **kwargs):
    """
    a bash ls imitation.
    allows wildcards
    """
    directory = resolve(directory)
    if os.path.isdir(directory):
        directory = os.path.join(directory, "*")
    return [os.path.basename(p) for p in glob.iglob(directory, **kwargs)]


def mkdir(path, isDir=False):
    """
    Notes:
        if you want to create a dir, add '/' behind path or use isDir=True
    Args:
        path (str): directory or path to create.
        isDir (str): tell mkdir that you are explicitly giving a directory.
    """
    log = logging.getLogger()
    path = resolve(path)
    if isDir:
        directory = path
    else:
        directory = os.path.dirname(path)
    if not os.path.exists(directory) and not directory == "":
        log.info("Create directory {0}".format(directory))
        os.makedirs(directory)
    else:
        log.warning("Directory {0} already exists".format(directory))


def dir_structure(directory):
    """
    Creates a nested dictionary that represents the folder structure of
    directory
    """
    dir_structure = {}
    directory = directory.rstrip(os.sep)
    start = directory.rfind(os.sep) + 1
    for path, dirs, files in os.walk(directory):
        folders = path[start:].split(os.sep)
        subdir = dict.fromkeys(files)
        parent = functools.reduce(dict.get, folders[:-1], dir_structure)
        parent[folders[-1]] = subdir
    if len(dir_structure) == 0:
        return {}
    return dir_structure[dir_structure.keys()[0]]  # first entry is always directory



"""
Publishing:
    either:
        $make publish
    or manually:
        run tests
        remember to change the version in setup(... ) below
        $ git commit -am "<my message>"
        $ git push
        $ git tag -a v<my.version.id> -m "<comment to my version>"  # tag version
        $ git push origin v<my.version.id>  # explicitly push tag to the shared server
        $ python setup.py sdist
        $ twine upload dist/*

"""
from setuptools import setup
setup()


# in the terminal:
# pip install .



from contextlib import contextmanager
import logging
import time


def progressbar(iterable, log=None):
    """
    Args:
        iterable
        log (logger): optional

    Examples:
        >>> import logging
        >>> import rna
        >>> import sys
        >>> sys.modules['tqdm'] = None
        >>> log = logging.getLogger(__name__)

        >>> a = range(3)
        >>> for value in rna.log.progressbar(a, log=log):
        ...     _ = value * 3

    """
    if log is None:
        log = logging.getLogger()
    try:
        from tqdm import tqdm as progressor

        tqdm_exists = True
    except ImportError:

        def progressor(iterable):
            """
            dummy function. Doe nothing
            """
            return iterable

        tqdm_exists = False
    try:
        nTotal = len(iterable)
    except Exception:
        nTotal = None

    for i in progressor(iterable):
        if not tqdm_exists:
            if nTotal is None:
                log.info("Progress: item {i}".format(**locals()))
            else:
                log.info("Progress: {i} / {nTotal}".format(**locals()))
        yield i


@contextmanager
def timeit(msg="No Description", log=None, precision=1) -> None:
    """
    Context manager for autmated timeing

    Args:
        msg (str): message to customize the log message
        log (logger)
        precision (int): show until 10^-<precision> digits
    """
    if log is None:
        log = logging.getLogger()
    startTime = time.time()
    log.log(logging.INFO, "-> " * 30)
    message = "Starting Process: {0} ->".format(msg)
    log.log(logging.INFO, message)

    yield

    log.log(
        logging.INFO,
        "\t\t\t\t\t\t<- Process Duration:"
        "{value:1.{precision}f} s".format(
            value=time.time() - startTime, precision=precision
        ),
    )
    log.log(logging.INFO, "<- " * 30)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Use pyqtgraph for fast 3d plots especially.
pyqtgraphs GFViewWidget is used as a correlate to matplotlibs Axes
"""
import rna
import numpy as np
import sys
import pyqtgraph as pg
import pyqtgraph.opengl
import pyqtgraph.exporters


pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'w')


def axes_dim(axes):
    return 3


global window, view
window, view = None, None


def gca(dim=None):
    if dim is None:
        dim = 3
    if dim != 3:
        raise NotImplementedError()
    global window, view
    if view is None:
        pg.mkQApp()
        window = pg.LayoutWidget()
        view = pg.opengl.GLViewWidget()
        view.setBackgroundColor('w')
        window.resize(800, 800)
    return view


def show():
    global window, view
    window.addWidget(view)
    window.show()

    if (sys.flags.interactive != 1) or not hasattr(pg.Qt.QtCore, 'PYQT_VERSION'):
        pg.Qt.QtGui.QApplication.instance().exec_()


def clear(axes):
    global window, view
    view.setParent(None)
    # window.removeWidget(view)
    view = None


def _save(path, **kwargs):
    axes = kwargs.get('axes', None)
    if axes is None:
        axes = gca()
    axes.grabFrameBuffer().save(path, **kwargs)
    # global window
    # exporter = pg.exporters.ImageExporter(window.plotItem)
    # exporter.parameters().update(kwargs)
    # exporter.export(path)


class GraphicsViewOverlay2D(pg.GraphicsView):
    """
    GraphicsView subclass that uses GLViewWidget as its canvas.
    This allows 2D graphics to be overlaid on a 3D background.
    """
    def __init__(self, axes):
        """
        Args:
            axes (gl.GLViewWidget): view upon which to copy all events
        """
        self.glView = axes
        pg.GraphicsView.__init__(self, background=None)
        self.setStyleSheet("background: transparent")
        self.setViewport(self.glView)

    def mousePos(self, event):
        """
        Distribute mouse events to both widgets
        """
        pg.GraphicsView.mousePos(self, event)
        self.glView.mousePos(event)

    def paintEvent(self, event):
        """
        Distribute paint events to both widgets
        """
        self.glView.paintEvent(event)
        return pg.GraphicsView.paintEvent(self, event)

    def mousePressEvent(self, event):
        """
        Distribute mouse events to both widgets
        """
        pg.GraphicsView.mousePressEvent(self, event)
        self.glView.mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """
        Distribute mouse events to both widgets
        """
        pg.GraphicsView.mouseMoveEvent(self, event)
        self.glView.mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        """
        Distribute mouse events to both widgets
        """
        pg.GraphicsView.mouseReleaseEvent(self, event)
        self.glView.mouseReleaseEvent(event)


def plot_array(*args, **kwargs):
    raise NotImplementedError()


def plot_mesh(vertices, faces, **kwargs):
    po = rna.plotting.PlotManager(kwargs)
    po.setdefault('drawEdges', False)
    po.setdefault('smooth', False)
    po.setdefault('shader', None)
    scalars = po.pop('color', None)
    if scalars is None:
        scalars = [0] * len(faces)
    cmap, vmin, vmax = po.get_norm_args(
        vmin_default=min(scalars),
        vmax_default=max(scalars),
        cmap_default='viridis')

    scalars = np.array(scalars)
    scalars[scalars < vmin] = vmin
    scalars[scalars > vmax] = vmax
    norms = (scalars - vmin) / (vmax - vmin)

    color_map = get_cmap(cmap)
    colors = color_map.map(norms, mode='float')

    mesh_data = pg.opengl.MeshData(
            vertexes=np.array(vertices),
            faces=np.array(faces),
            faceColors=colors)
    artist = pg.opengl.GLMeshItem(
        meshdata=mesh_data,
        **po.plot_kwargs)
    artist.colors = colors
    artist.vmin = vmin
    artist.vmax = vmax
    artist.color_map = color_map
    artist.setGLOptions('opaque')

    po.axes.addItem(artist)
    return artist


def get_cmap(cm_name):
    """
    Use matplotlib conventions to retrieve a pg.ColorMap

    Examples:
        >>> import rna
        >>> rna.plotting.use('pyqtgraph')
        >>> cm = rna.plotting.backend.get_cmap('viridis')
    """
    import matplotlib.pyplot as plt
    pltMap = plt.get_cmap(cm_name)
    colors = pltMap.colors
    colors = [c + [1.] for c in colors]
    positions = np.linspace(0, 1, len(colors))
    cmap = pg.ColorMap(positions, colors)
    return cmap


def set_colorbar(axes, artist, label=None, position='right',
                 **kwargs):
    # Make a 2D color bar using the same ColorMap
    overlay_view = GraphicsViewOverlay2D(axes)
    kwargs['size'] = kwargs.pop('size', (20, 200))
    kwargs['offset'] = kwargs.pop('offset', (15, -25))
    colorbar_legend = Colorbar(artist, label=label, **kwargs)
    overlay_view.addItem(colorbar_legend)
    return overlay_view


class Colorbar(pg.GradientLegend):
    def __init__(self, artist, *args, **kwargs):
        self.label = kwargs.pop('label')
        self.format = kwargs.pop('format', '{0:.2f}')
        self.vmin, self.vmax = artist.vmin, artist.vmax
        ticks = kwargs.pop('ticks', np.linspace(self.vmin, self.vmax, 4))
        super(Colorbar, self).__init__(*args, **kwargs)
        color_map = artist.color_map
        self.setGradient(color_map.getGradient())
        self.set_tick_positions(ticks)

    def set_tick_positions(self, tick_positions):
        if self.vmin > min(tick_positions) or self.vmax < max(tick_positions):
            raise ValueError("tick positions must lie in (vmin, vmax)")

        self.setLabels(dict([
            (self.format.format(v),
             1 / (self.vmax - self.vmin) * v - self.vmin / (self.vmax - self.vmin))  # NOQA
            for v in tick_positions
        ]))

    def paint(self, p, opt, widget):
        pg.UIGraphicsItem.paint(self, p, opt, widget)
        rect = self.boundingRect()  # Bounds of visible area in scene coords.
        unit = self.pixelSize()  # Size of one view pixel in scene coords.
        if unit[0] is None:
            return

        # determine max width of all labels
        labelWidth = 0
        labelHeight = 0
        for k in self.labels:
            b = p.boundingRect(
                pg.Qt.QtCore.QRectF(0, 0, 0, 0),
                pg.Qt.QtCore.Qt.AlignLeft | pg.Qt.QtCore.Qt.AlignVCenter,
                str(k))
            labelWidth = max(labelWidth, b.width())
            labelHeight = max(labelHeight, b.height())

        b_tick = p.boundingRect(
            pg.Qt.QtCore.QRectF(0, 0, 0, 0),
            pg.Qt.QtCore.Qt.AlignCenter | pg.Qt.QtCore.Qt.AlignVCenter,
            '-')
        b_label = p.boundingRect(
            pg.Qt.QtCore.QRectF(0, 0, 0, 0),
            pg.Qt.QtCore.Qt.AlignCenter | pg.Qt.QtCore.Qt.AlignVCenter,
            str(self.label))

        labelWidth *= unit[0]
        labelHeight *= unit[1]

        textPadding = 2  # in px

        if self.offset[0] < 0:
            x3 = rect.right() + unit[0] * self.offset[0]
            x2 = x3 - labelWidth - unit[0]*textPadding*2
            x1 = x2 - unit[0] * self.size[0]
        else:
            x1 = rect.left() + unit[0] * self.offset[0]
            x2 = x1 + unit[0] * self.size[0]
            x3 = x2 + labelWidth + unit[0]*textPadding*2
        if self.offset[1] < 0:
            y2 = rect.top() - unit[1] * self.offset[1]
            y1 = y2 + unit[1] * self.size[1]
        else:
            y1 = rect.bottom() - unit[1] * self.offset[1]
            y2 = y1 - unit[1] * self.size[1]
        self.b = [x1, x2, x3, y1, y2, labelWidth]

        # Have to scale painter so that text and gradients are correct size.
        p.scale(unit[0], unit[1])

        # Draw color bar
        self.gradient.setStart(0, y1/unit[1])
        self.gradient.setFinalStop(0, y2/unit[1])
        p.setBrush(self.gradient)
        rect = pg.Qt.QtCore.QRectF(
            pg.Qt.QtCore.QPointF(x1/unit[0], y1/unit[1]),
            pg.Qt.QtCore.QPointF(x2/unit[0], y2/unit[1])
        )
        p.drawRect(rect)

        # draw labels
        p.setPen(pg.Qt.QtGui.QPen(pg.Qt.QtGui.QColor(0, 0, 0)))
        tx = x2 + unit[0]*textPadding
        lh = labelHeight/unit[1]
        for k in self.labels:
            y = y1 + self.labels[k] * (y2-y1)
            p.drawText(
                pg.Qt.QtCore.QRectF(tx/unit[0] + b_tick.width(),
                                    y/unit[1] - lh/2.0,
                                    1000,
                                    lh),
                pg.Qt.QtCore.Qt.AlignLeft | pg.Qt.QtCore.Qt.AlignVCenter,
                str(k))
            p.drawText(
                pg.Qt.QtCore.QRectF(x2,
                                    y/unit[1] - b_tick.height()/2.0 - 2,  # the -2 is a hack to move the ticks up  # NOQA
                                    b_tick.width(),
                                    b_tick.height()),
                pg.Qt.QtCore.Qt.AlignLeft | pg.Qt.QtCore.Qt.AlignVCenter,
                '-')
        p.save()
        p.translate(x3 + b_label.height(),
                    y1 + 0.5 * (y2 - y1) - b_label.width() / 2)
        p.rotate(90)
        p.drawText(0., 0., str(self.label))
        p.restore()


if __name__ == '__main__':
    import doctest
    doctest.testmod()
